/// \file WHAIDemo.cpp

#include <fezoolib/Tracking/WHAITracker.hpp>
#include <fezoolib/Core/ToolConfig.hpp>
#include <fezoolib/Core/Timestamp.hpp>
#include <fezoolib/Core/DepthGeometry.hpp>
#include <boost/thread/thread.hpp>

#include <signal.h>
#include<dirent.h>

bool shutdown = false;

//#define AVOID_QT //AGIL: To make it work on debian

namespace WHAIConf
{
// Tool name
static const std::string TOOL_NAME   = "WHAI Tracker Demo";
// Tool brief
static const std::string BRIEF       = "Track hands easily with WHAI Tracker.\n";
// Tool description
static const std::string DESCRIPTION = (std::string) "The tracker will run in real-time from an OpenNI compatible camera.\n"+
		"If a test_sequence (.oni) is provided, it will run offline.\n" +
		"Config file is in whai.cfg. \n"+
		"This file is important, because it sets the threshold in order to init tracks\n" +

		"\nHave a WHAI tracking!\n\n"+
		"\n                        The Exipple team.\n"+
		"\n==============================================================================\n";
}

using namespace WHAIConf;

class TestConfig : public gestoos::ToolConfig
{
public:
	std::string test_sequence;
	std::string cfg_file;
	std::string resources_path;
	bool gpu;


	TestConfig(int argc, char* argv[]) :
		gestoos::ToolConfig(TOOL_NAME, BRIEF, DESCRIPTION)
	{
		test_sequence="";
		cfg_file = "config/whai.cfg";
#ifdef MSVC
        resources_path = "/imatge/acasanova/workspace/GestoosSDK-0.8.1/data/";
#else
        resources_path = "/imatge/acasanova/workspace/GestoosSDK-0.8.1/data/";
#endif
		gpu = false;
//#ifdef MSVC
		cfg_file = "/imatge/acasanova/workspace/GestoosSDK-0.8.1/samples/WHAIDemo/config/whai.cfg";
//#endif
		//add options
		add_option(test_sequence, "test_sequence", "Full path of the test sequence, in oni format. If not provided, hand tracker runs on the camera.");
		add_option(cfg_file, "cfg_file", "Configuration file for WHAI Tracker (whai.cfg).");
        add_option(resources_path, "resources_path", "path to data models folder");
		add_flag(gpu, "gpu", "use gpu computation ", 'g');

		// configure
		configure(argc, argv);

		//read
		read("test_sequence", test_sequence);
		read("cfg_file", cfg_file);
		read("resources_path", resources_path);
		read("gpu", gpu);

	}

};

//Test on a live device
gestoos::CaptureRGBD capture;

#ifndef MSVC
struct sigaction sa;
#endif

static void got_signal(int s)
{
	shutdown = true;

	std::cout << "SHUTDOWN!!!!!!!" << std::endl;
	std::cout<<"\nstopping camera\n"<<std::flush;
	capture.stop();

	std::cout<<"exit(0)\n"<<std::flush;
	//exit(0);
}


/// Renderinf function called by the render thread
void render_func(const cv::Mat & depth_map, gestoos::tracking::WHAITracker * tracker, std::vector <cv::Point3f> * detected_coords)
//void render_func(const cv::Mat & depth_map, gestoos::tracking::WHAITracker * tracker)
{
#ifndef AVOID_QT
	//Exit if empty image!
	if( depth_map.cols == 0 )
	{
		return;
	}

	cv::Mat color_img;

#ifndef NDEBUG
	std::cout << "DBG: Depth map conversion \n" << std::flush;
#endif

	// Normalizing depth data values
	depth_map.convertTo(color_img, CV_8UC1, 1/15.0);
	// Converting from Grayscale to RGB
    cv::cvtColor(color_img, color_img, CV_GRAY2BGR);

    //Set to false to hide trajectory
	bool show_trajectory = true;

	// Comment to hide 2D coordinates

	std::vector< gestoos::tracking::ObjectTrack *> objectTrack = tracker->active_tracks();

	if (!objectTrack.empty() && objectTrack.size()==1){
		//if (detected_coords->size()<100){
	gestoos::tracking::ObjectTrack::point_type coord = objectTrack[0]->get_position();
	std::cout << "Coord  num "<<detected_coords->size()<<" (" << coord.x << " " << coord.y << " " << coord.z << ")" << std::endl;
   detected_coords->push_back(coord);
		//}
	}


	//Comment to hide hand position and trajectory
	tracker->visualize(color_img, 1, show_trajectory);
	//Comment to hide hand labels
	tracker->show_labels(color_img, 1);

#ifndef NDEBUG
	std::cout << "DBG: Drawing... \n" << std::flush;
#endif

	cv::imshow("Tracker Output", color_img);
//	std::string storedName="/imatge/acasanova/work/090516Capture/03-fingers-NR/01kinect/detectedFingers/Coord"+std::to_string(detected_coords->size())+"detected.jpg";
//	cv::imwrite(storedName,color_img);

#ifndef NDEBUG
	std::cout << "DBG: Done \n" << std::flush;
#endif

	// pthreads and imshow don't get along in Apple machines
#ifndef __APPLE__
#ifndef MSVC

	int key = cv::waitKey(1);
	if (key==27)
	{
		shutdown = true;
		// break;
	}
#endif
#endif

#endif //AVOID_QT



}


int main(int argc, char* argv[])
{

	const std::string inputSettingsFile = argc > 1 ? argv[1] : "default.xml";
	std::string oni_file,img_dir,output_file;
		 cv::FileStorage m(inputSettingsFile,cv::FileStorage::READ);

					    if (!m.isOpened())
					        std::cout << "Could not open Client configuration file: \\" << std::endl;

					    m["oni_file"] >> oni_file;
					    m["img_dir"] >> img_dir;
					    m["output_file"] >> output_file;

					    m.release();

	shutdown = false;

	//TestConfig cfg(argc, argv);

	//Declare WHAITracker
	gestoos::tracking::WHAITracker whai;

	//To store the current frame num
	gestoos::tracking::ObjectTrack::ts_type frame=0;

	//To store detected coordinates in XML file
		std::vector <cv::Point3f> detected_coords;


/*	if (cfg.test_sequence!="")
	{
		capture.init(cfg.test_sequence);
	}
	else
	{*/
		bool camera_ready = capture.init("", gestoos::CaptureRGBD::SensorMode::OPENNI, gestoos::CaptureRGBD::FPS30, gestoos::CaptureRGBD::QVGA); //, true, false);
		//bool camera_ready = capture.init(oni_file, gestoos::CaptureRGBD::SensorMode::OPENNI, gestoos::CaptureRGBD::FPS30, gestoos::CaptureRGBD::QVGA);
		if (!camera_ready)
		{
			exit(-1);
		}
//	}

	/*if(cfg.gpu)
	{
		std::cout << "using gpu" << std::endl;
	}
	*/
	whai.set_video_mode(gestoos::CaptureRGBD::QVGA);

	//Configure the tracker
	whai.init("/imatge/acasanova/workspace/GestoosSDK-0.8.1/samples/WHAIDemo/config/whai.cfg");




	/*
	 * Tracking inputs
	 */
	cv::Mat depth_map;
	cv::Mat color_img;
#ifndef AVOID_QT
	cv::namedWindow("Tracker Output",cv::WINDOW_NORMAL);
#endif

	signal(SIGINT, got_signal);
/*
	std::string dirName = img_dir;
	std::vector<std::string> img_list;
			DIR *dir;
			dir = opendir(dirName.c_str());
			std::string imgName;
			struct dirent *ent;
			if (dir != NULL) {
			while ((ent = readdir (dir)) != NULL) {
			imgName= ent->d_name;
			img_list.push_back(dirName+imgName);
			}
			}
			std::sort(img_list.begin(),img_list.end());

int index_img=0;
*/
	while(!shutdown)
	{
		/*if(index_img<img_list.size()){

	   while(img_list[index_img].find(".png") == std::string::npos)
				index_img++;

		std::cout << img_list[index_img]<<std::endl;
		depth_map = cv::imread(img_list[index_img],cv::IMREAD_ANYDEPTH);
		index_img++;
		}

*/
		gestoos::Timestamp ts = gestoos::clock_ts();

		/*
		 * Capture
		 */

		capture.get_frame();

		depth_map = capture.depth_frame();
std::cout << "després d'agafar frame"<<std::endl;

		if (depth_map.empty())
		{
			std::cout << "Error: RGBD Sensor not found or disconnected " << std::endl;
			exit(-1);
		}


		/*
		 * Filter data, mainly filling depth holes
		 */
		depth_map=gestoos::depth_error_filter(depth_map);

		/*
		 * Detect hands using WHAI
		 */
		whai.update(depth_map, frame);

		//Launch render thread for 1 frame
		//Note: This works flawlessly on Ubuntu, but not on OSX:
		//http://answers.opencv.org/question/30507/imshow-not-working-in-pthread-macosx-109/
#ifndef AVOID_QT
#ifndef MSVC
		//boost::thread render_thread(render_func, depth_map, &whai,&detected_coords);
		render_func(depth_map,  &whai,&detected_coords);
		int key = cv::waitKey(1);
		if (key==27)
		{
			shutdown = true;
			// break;
		}
#endif

#ifdef __APPLE__
		int key = cv::waitKey(1);
		if (key==27)
		{
			shutdown = true;
			// break;
		}
#endif

#ifdef MSVC
		render_func(depth_map,  &whai,&detected_coords);
		int key = cv::waitKey(1);
		if (key==27)
		{
			shutdown = true;
			// break;
		}
#endif
#else //AVOID_QT
		std::vector<gestoos::tracking::ObjectTrack*> active_tracks = whai.active_tracks();
		std::vector<gestoos::tracking::ObjectTrack*>::const_iterator iter;
		for (iter = active_tracks.begin(); iter!=active_tracks.end(); ++iter)
		{
			std::cout << "Hand " << (*iter)->get_id() << " " << (*iter)->get_position_2d() << std::endl;
		}
#endif //AVOID_QT


		if(shutdown) break;
		++frame;


		gestoos::Timestamp ts2 = gestoos::clock_ts();

		gestoos::Timestamp diff_ts=ts2-ts;


		double elapsed= diff_ts.partial_seconds()*1000.0 + diff_ts.partial_milliseconds();


		std::cout << "\r" << elapsed << " ms | "<< 1/(elapsed/1000.0) << " fps" <<  std::flush;

	}

	//	closedir (dir);

cv::FileStorage fs(output_file, cv::FileStorage::WRITE);
	fs << "Detected_coords" << detected_coords;
	fs.release();
	std::cout << "Detected fingers stored."<<std::flush;


	return 0;
}




