#include <deque>

#include <fezoolib/Detection/GestureDetector.hpp>
#include <fezoolib/Tracking/ObjectTracker.hpp>
#include <fezoolib/Tracking/WHAITracker.hpp>
#include <fezoolib/Core/ToolConfig.hpp>
#include <fezoolib/Core/Timestamp.hpp>
#include <fezoolib/Core/DepthGeometry.hpp>
#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <signal.h>

#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/features2d.hpp>

#include <iostream>
#include <string>

#include<dirent.h>



#include "/usr/include/oscpack/osc/OscOutboundPacketStream.h"
#include "/usr/include/oscpack/ip/UdpSocket.h"


#define OUTPUT_BUFFER_SIZE 1024


cv::Mat heat_prob;
gestoos::CaptureRGBD capture;

float resolutionX, resolutionY;
float xzFactor, yzFactor;

gestoos::tracking::WHAITracker detector;
gestoos::tracking::ObjectTracker tracker;
double _init_threshold;
double _tracking_threshold;
std::deque<gestoos::tracking::ObjectTrack> candidates;
std::deque<cv::Point3f> candidates_pres;
boost::mutex mutex, mutex_pres;

std::string img_dir;
std::string server_ip;
std::string oni_file;
std::string ini_file;
int client_id;
int server_port;
int ini_frame;

bool can_send_tracks = false;

bool real_world_coords = false;


namespace gestoos_client
{
// Tool name
static const std::string TOOL_NAME   = "gestoos_client";
// Tool brief
static const std::string BRIEF       = "A TCP Client that delivers finger detection candidates to TCP server \n";
// Tool description
static const std::string DESCRIPTION = (std::string) "\n";
}

using namespace gestoos_client;
using namespace gestoos;

class TestConfig : public gestoos::ToolConfig
{
public:

	std::string img_dir;
	std::string server_ip;
	std::string oni_file;
	std::string ini_file;
	int client_id;
	int server_port;

	TestConfig(int argc, char* argv[]) :
		gestoos::ToolConfig(TOOL_NAME, BRIEF, DESCRIPTION)
	{

		oni_file = "";
		ini_file ="./config/tracker.ini";
		client_id = 0;
		server_ip = "127.0.0.1";
		server_port = 23424;

#ifdef MSVC
		ini_file ="./../../config/trackerWin.ini";
#endif

		//add options
		add_option(oni_file, "oni_file", "Provide the path to a .oni to run the client from recorded data");
		add_option(ini_file, "ini_file", "Configuration file with gesture and thresholds");
		add_option(client_id, "client_id", "Client ID");
		add_option(server_ip, "server_ip", "connection ip", 'i');
		add_option(server_port, "server_port", "connection port", 'p');
		// configure
		configure(argc, argv);

		//read
		read("server_ip", server_ip);
		read("client_id", client_id);
		read("server_port", server_port);
		read("ini_file", ini_file);
		read("oni_file", oni_file);

	}

};

bool is_processing=false;
//Forward declaration
void got_signal(int);

bool global_detect = true;

class ThreadedTracker
{
public:
	ThreadedTracker()
	{
		detector = NULL;
		tracker  = NULL;
		capture  = NULL;
	//	whai = NULL;
	}

	~ThreadedTracker()
	{}

	std::vector<cv::Point3d> flood_fill_up( const cv::Mat & depth_map, const cv::Point2d & root)
					{
		std::vector<cv::Point3d> ret;
		float accum_dist;



		unsigned short root_d = depth_map.at<unsigned short>(root.y, root.x);

		std::deque<cv::Point2d> S;
		std::vector<cv::Point2d> visited;

		S.push_back( root );
		visited.push_back(root);

		while( !S.empty() )
		{

			cv::Point2d pt = S.front();
			S.pop_front();

			unsigned short pt_z = depth_map.at<unsigned short>(pt.y, pt.x);


			if( pt.x > 0  				&& std::find(visited.begin(), visited.end(), pt + cv::Point2d(-1, 0))==visited.end() )
			{
				S.push_back( pt + cv::Point2d(-1, 0));
				visited.push_back( pt + cv::Point2d(-1, 0));
			}
			if( pt.x < depth_map.cols-1 && std::find(visited.begin(), visited.end(), pt + cv::Point2d( 1, 0))==visited.end() )
			{
				S.push_back( pt + cv::Point2d( 1, 0));
				visited.push_back( pt + cv::Point2d( 1, 0));
			}
			if( pt.y > 0 				&& std::find(visited.begin(), visited.end(), pt + cv::Point2d( 0, 1))==visited.end() )
			{
				S.push_back( pt + cv::Point2d( 0, 1));
				visited.push_back( pt + cv::Point2d( 0, 1));
			}

			ret.push_back(cv::Point3d(pt.x, pt.y, pt_z));

		}
		return ret;
					}


	void operator()()
	{
	cv::namedWindow("Tracker Output",cv::WINDOW_NORMAL);
		cv::moveWindow("Tracker Output", 10, 50);
		if (detector!=NULL)
		{

			frame = 0;
			cv::Mat color_img;

			while(global_detect)
			{
				//std::cout << "detecting " << std::endl;
				/*
				 * Capture
				 */
				capture->get_depth_frame();
			    cv:: Mat depth_map = capture->depth_frame();

				if (depth_map.empty())
				{
					std::cout << "Error: RGBD Sensor not found or disconnected " << std::endl;
					exit(-1);
				}
				/*
				 * Filter data, mainly filling depth holes
				 */
				depth_map=gestoos::depth_error_filter(depth_map);

				detector->update(depth_map, frame);
			/*
				 *  Do detection
				 */
				depth_map.convertTo(color_img, CV_8UC1, 1/15.0);
				std::deque<gestoos::tracking::ObjectTrack>  cand;
				std::vector< gestoos::tracking::ObjectTrack *> objectTrack = detector->active_tracks();

					if (!objectTrack.empty() && objectTrack.size()==1){
						cand.push_back(*objectTrack[0]);
						gestoos::tracking::ObjectTrack::point_type coord = objectTrack[0]->get_position();
					}

				//Comment to hide hand position and trajectory
				detector->visualize(color_img, 1, true);
				//Comment to hide hand labels
				detector->show_labels(color_img, 1);
				cv::imshow("Tracker Output", color_img);
				int key = cv::waitKey(1);
				if (key==27)
				{

					 //break;
				}


				//Thread safe append to candidates
				{
					boost::mutex::scoped_lock scoped_lock(*mutex);
					//					candidates.resize(cand.size());
					candidates.insert(candidates.end(), cand.begin(), cand.end());
				}



			//	cv::imshow("Tracker Output", color_img);

			//	int key = cv::waitKey(1);
			}
		}

	}


	cv::Mat depth_map, detector_map;
	std::string ini_file;
	//Test on a live device
	gestoos::CaptureRGBD * capture;

	//Declare WHAITracker
	gestoos::tracking::WHAITracker * detector;
	gestoos::tracking::ObjectTrack::ts_type frame;

	//gestoos::detection::GestureDetector * detector;
	gestoos::tracking::ObjectTracker * tracker;
	//gestoos::tracking::WHAITracker * wtracker;
	boost::mutex *	mutex;
	boost::mutex *	mutex_pres;

};



bool quit = false;    // signal flag

void got_signal(int)
{
	std::cout<<"Closing sensor"<<std::endl;
	capture.stop();

	std::cout<<"exit(0)"<<std::endl;
	exit(0);
}

void convertDepthToWorldCoordinates(float depthX, float depthY, float depthZ, float *realX, float *realY, float *realZ) {

	double cx_ir=1;
	double cy_ir=1;
	double fx_ir=1;
	double fy_ir=1;

	   *realZ=depthZ/1000;
	   *realX = (float) (((depthX- cx_ir) * (*realZ))/ fx_ir);
	   *realY = (float) (((depthY- cy_ir) * (*realZ))/ fy_ir);
}


void runClient( int client_id, const std::string & ip, int port) {

	UdpTransmitSocket transmitSocket( IpEndpointName( ip.c_str(), port ) );



	while( true )
	{



		//Thread safe copy
		std::deque<gestoos::tracking::ObjectTrack> cand;
		{
			boost::mutex::scoped_lock scoped_lock(mutex);
			cand.resize( candidates.size() );
			std::copy( candidates.begin(), candidates.end(), cand.begin());
			candidates.clear();
		}




		char buffer[OUTPUT_BUFFER_SIZE];
		osc::OutboundPacketStream p( buffer, OUTPUT_BUFFER_SIZE );

		if( cand.empty())
		{
			usleep(20*1000);
			continue;
		}
		else if (!cand.empty())
		{
			// Message for hand candidates
			p << osc::BeginMessage( "/Cand" ) << (osc::int32)client_id << (osc::int32)cand.size() ;

			int i = 0;
			for( auto it=cand.begin(); it!=cand.end(); ++it, ++i)
			{

				float xCoord, yCoord, zCoord;

				if (real_world_coords)
				{
					convertDepthToWorldCoordinates(it->get_position().x, it->get_position().y, it->get_position().z, &xCoord, &yCoord, &zCoord);
				}
				else
				{

					xCoord = it->get_position().x;
					yCoord = it->get_position().y;
					zCoord = it->get_position().z;

				}

				//Add to message
				p << xCoord << yCoord << zCoord << (float)it->get_score() ;
				std::cout << "Sent message" <<xCoord <<" "<< yCoord << " "<<zCoord << (float)it->get_score() ;
			}

			p << osc::EndMessage;
			std::cout << "Data send :"<<p.Data()<<std::endl;
			transmitSocket.Send( p.Data(), p.Size() );
			p.Clear();
		}


	}

}


int main(int argc, char* argv[])
{
	const std::string inputSettingsFile = argc > 1 ? argv[1] : "default.xml";

	 cv::FileStorage m(inputSettingsFile,cv::FileStorage::READ);

				    if (!m.isOpened())
				        std::cout << "Could not open Client configuration file: \\" << std::endl;

				    m["oni_file"] >> oni_file;
				    m["img_dir"] >> img_dir;
				    m["client_id"] >> client_id;
				    m["server_ip"] >> server_ip;
				    m["server_port"] >> server_port;
				    m["ini_frame"] >> ini_frame;
				    m.release();

	bool camera_ready=false;
	if (oni_file=="")
	{
		camera_ready=capture.init("", gestoos::CaptureRGBD::SensorMode::OPENNI, gestoos::CaptureRGBD::FPS30, gestoos::CaptureRGBD::QVGA); //, true, false);
	}
	else
		camera_ready=capture.init(oni_file, gestoos::CaptureRGBD::SensorMode::OPENNI, gestoos::CaptureRGBD::FPS30, gestoos::CaptureRGBD::QVGA); //, true, false);


#ifdef USE_GLSL
	detector.use_gpu_detection(use_gpu);
#endif

	detector.init("WHAIDemo_client/config/whai.cfg");



	ThreadedTracker thread_tracker;

	boost::thread bt_tracker;


	thread_tracker.capture 		= &capture;
	thread_tracker.detector 	= &detector;
	thread_tracker.tracker 		= &tracker;
	thread_tracker.mutex 		= &mutex;
	thread_tracker.mutex_pres	= &mutex_pres;


	std::cout << "Launching tracker thread ... \n" << std::flush;
	bt_tracker = boost::thread(thread_tracker);

	runClient(client_id, server_ip, server_port);

	bt_tracker.join();
	return 0;

}
