#include <iostream>
#include <fstream>      // std::ifstream


#include <fezoolib/Core/ToolConfig.hpp>

#include <GestureComposer.hpp>
#include <cmath>

using namespace gestoos;
using namespace std;

class TestConfig : public gestoos::ToolConfig
{
public:

	std::string render_ip;
	int render_port;
	bool calibrate;

	TestConfig(int argc, char* argv[]) :
		gestoos::ToolConfig("mod hand tracker", " hand tracker", "none")
	{

		calibrate = false;

		//add options
		add_option(render_ip, "render_ip", "Provide the IP address to the rendering server", 'i');
		add_option(render_port, "render_port", "Connection port of the rendering server", 'p');
		add_flag(calibrate, "calibrate", "Run in calibration mode", 'c');

		// configure
		configure(argc, argv);

		//read
		read("render_ip", render_ip);
		read("render_port", render_port);
		read("calibrate", calibrate);
	}

};


class ExamplePacketListener : public osc::OscPacketListener {
public:

	GestureComposer * gc;
	gestoos::Timestamp _last_received;
	cv::Mat TF;

protected:

	virtual void ProcessMessage( const osc::ReceivedMessage& m,
			const IpEndpointName& remoteEndpoint )
	{
		(void) remoteEndpoint; // suppress unused parameter warning

		try{

				osc::ReceivedMessageArgumentStream args = m.ArgumentStream();

				//std::string ts;d
				osc::int32 num, id;
				float  x, y, z, score;

				args >> id >> num;


				std::vector<Candidate> new_cand;
				std::vector<cv::Point3f> new_cand_d;
				std::vector<int> client_list;
				float realX,realY,realZ;
				cv::Point3f point_RT;

					args >>  x >> y >> z >> score ;
					std::cout << "Received message from "<< id << ": " <<x << " " << y << " " << z<<std::endl;



					if (id==1){
						convertDepthToWorldCoordinates(x,y,z,&realX,&realY,&realZ,1);

						applyRT(cv::Point3f(realX,realY,realZ),&point_RT);
						realX=point_RT.x;
						realY=point_RT.y;
						realZ=point_RT.z;
					}else{
						convertDepthToWorldCoordinates(x,y,z,&realX,&realY,&realZ,5);

					}
						long int my_stamp = gestoos::clock_ts().total_milliseconds();


					std::cout << "Point stored as candidate(TS " <<my_stamp<< ") : "<< realX<< " "<<realY << " " << realZ<<std::endl;
					Candidate cand( my_stamp, realX, realY, realZ, score, gc->get_merge_radius(), 0);
					new_cand.push_back( cand );
					new_cand_d.push_back( cv::Point3f(realX,realY,realZ));
					client_list.push_back( id);	// Add candidate
				//}
				args >> osc::EndMessage;

				//Thread safe copy
				//if( gc->RT.count(id) )
				//{

				{
				boost::mutex::scoped_lock scoped_lock(gc->mutex);
					gc->candidates.insert(gc->candidates.begin(), new_cand.begin(), new_cand.end());
					gc->clients.insert(gc->clients.begin(), client_list.begin(), client_list.end());
					gc->candidates_d.insert(gc->candidates_d.begin(), new_cand_d.begin(), new_cand_d.end());

				}


					}catch( osc::Exception& e ){
			// any parsing errors such as unexpected argument types, or
			// missing arguments get thrown as exceptions.
			std::cout << "error while parsing message: "
					<< m.AddressPattern() << ": " << e.what() << "\n";
		}
	}
	void applyRT(cv::Point3f point, cv::Point3f *point_RT)
	{
cv::Mat point_m= cv::Mat::zeros (4,1,5);
point_m.at<float>(0,0)=point.x;
point_m.at<float>(1,0)=point.y;
point_m.at<float>(2,0)=point.z;
point_m.at<float>(3,0)=1;

TF.convertTo(TF,CV_32F);

cv::Mat point_RT_m= TF*point_m;
point_RT_m=point_RT_m/point_RT_m.at<float>(3,0);
point_RT->x=point_RT_m.at<float>(0,0);
point_RT->y=point_RT_m.at<float>(1,0);
point_RT->z=point_RT_m.at<float>(2,0);


	}
void convertDepthToWorldCoordinates(float depthX, float depthY, float depthZ, float *realX, float *realY, float *realZ, int cam) {

		double cx_ir,cy_ir,fx_ir,fy_ir;

	if (cam==1){

	cx_ir=317.26;
	cy_ir=240.12;
	fx_ir=575.31;
	fy_ir=577.32;

	} else{

	cx_ir=315.75;
	cy_ir=249.61;
	fx_ir=576.98;
	fy_ir=579.75;

	}

		   *realZ=depthZ/1000;
		   *realX = (float) (((depthX- cx_ir) * (*realZ))/ fx_ir);
		   *realY = (float) (((depthY- cy_ir) * (*realZ))/ fy_ir);
	}
};


int main(int argc, char* argv[])
{
	//TestConfig cfg(argc, argv);
	const std::string inputSettingsFile = argc > 1 ? argv[1] : "default.xml";
cv::Mat TF;
		 cv::FileStorage m(inputSettingsFile,cv::FileStorage::READ);

					    if (!m.isOpened())
					        std::cout << "Could not open Client configuration file: \\" << std::endl;

					    m["FinalTF"] >> TF;

					    m.release();


	GestureComposer composer("192.168.1.202", 23422);



	//Start tracker thread
	ThreadedTracker threaded_tracker;
	threaded_tracker.gc = &composer;
	boost::thread bt(threaded_tracker);
	std::cout<<"Tracker created."<<std::endl;

	ExamplePacketListener listener;

	listener._last_received=0;
	listener.gc = &composer;
	listener.TF=TF;
	UdpListeningReceiveSocket s(
			IpEndpointName( "192.168.1.202", 23422 ),
			&listener );
	std::cout<<"UDP listener created."<<std::endl;

	s.RunUntilSigInt();


	std::cout << "detections reached!!"<<std::endl;
	std::cout << threaded_tracker.gc->detections.size()<<std::endl;
	cv::FileStorage fs("/imatge/acasanova/work/test.xml",
	cv::FileStorage::WRITE);

	if (!fs.isOpened())
		cout << "Could not open the configuration file: \\" << endl;
	std::vector<cv::Point3f> vec=threaded_tracker.gc->detections;
	std::vector<int> vec2=threaded_tracker.gc->stored_clients;
	fs << "Detections" << vec;
	fs << "Clients" << vec2;
	fs.release();

	std::cout<<"Closed UdpListeningReceiveSocket, ctrl+c again to exit"<<std::endl;
	bt.join();

	return 0;
}
