/*
 * GestureComposer.hpp
 *
 *  Created on: Apr 27, 2015
 *      Author: fezoo
 *
 * Modified on: May 15,2016
 * 		Author: Arantxa Casanova
 */

#ifndef SAMPLESSDK_GESTOOS_CLIENT_SRC_GESTURECOMPOSER_HPP_
#define SAMPLESSDK_GESTOOS_CLIENT_SRC_GESTURECOMPOSER_HPP_

#include <iostream>
#include <fstream>
#include <cstring>

#if defined(__BORLANDC__) // workaround for BCB4 release build intrinsics bug
namespace std {
using ::__strcmp__;  // avoid error: E2316 '__strcmp__' is not a member of 'std'.
}
#endif

#include <fezoolib/Core/Timestamp.hpp>
#include <fezoolib/Tracking/ObjectTracker.hpp>
#include <fezoolib/Tracking/Point3dTracker.hpp>


#include <opencv2/core.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>


#include <boost/format.hpp>
#include <boost/thread.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/thread/mutex.hpp>

#include "/usr/include/oscpack/osc/OscOutboundPacketStream.h"
#include "/usr/include/oscpack/ip/UdpSocket.h"
#include "/usr/include/oscpack/osc/OscReceivedElements.h"
#include "/usr/include/oscpack/osc/OscPacketListener.h"



const int GESTOOS_PORT = 1899;


static gestoos::tracking::TrackerParams tracker_params;

class ShowLabel {
public:

	ShowLabel(const std::string & _t, const cv::Point2f & _p, const cv::Scalar & _c) : text(_t), pos(_p), color(_c)
	{
		count = 0;
		count_max = 30;
		dead = false;
	}

	~ShowLabel(){}

	void update()
	{
		if( dead ) return;
		pos.y -= 5;
		count++;
		if( count == count_max ) dead = true;
	}

	bool is_dead()
	{
		return dead;
	}

	void draw(cv::Mat & img)
	{
		float fade = ( count_max - count ) / (float)count_max;
		cv::putText(img, text , pos,  cv::FONT_HERSHEY_SIMPLEX, 0.5 , color * fade , 1, 16);
	}

private:

	cv::Point2f pos;
	std::string text;
	int count, count_max;
	bool dead;
	cv::Scalar color;
};

class Candidate : public gestoos::tracking::ObjectTrack
{
public:
	Candidate():  gestoos::tracking::ObjectTrack() {}

	Candidate(long int ts, float x, float y, float z, double sc, double rad, int id) :  gestoos::tracking::ObjectTrack()
	{

		this->init(cv::Point3f(x,y,z),rad,sc,ts,tracker_params);
		this->set_id(id);

		//		sock = sk;
		client_id = id;
		timestamp = ts;
	}

	~Candidate()
	{}

	void print()
	{
		std::cout<<"["<<client_id<<"] "<< get_pt_world()<<" : "<<get_score()<<" : ts="<<timestamp<<std::endl;
	}


	int client_id;
//	SockAddr sock;
	gestoos::Timestamp timestamp;
};




inline void elapsed_time_since( const gestoos::Timestamp & t, const std::string & txt)
{
	gestoos::Timestamp toc = gestoos::clock_ts();
	gestoos::Timestamp diff_ts=toc-t;
	double elapsed= diff_ts.partial_seconds()*1000.0 + diff_ts.partial_milliseconds();

	//std::cout<<txt<<" time: "<<elapsed<<" ms"<<std::endl;
}


/// Main hand event processor class
class GestureComposer
{
public:
	/// Empty constructor
	GestureComposer(const std::string& ip, int port);
	/// Destructor
	~GestureComposer();
	void update();
	void print();
	void send_hand_positions();
	void send_presence_positions();


	double get_merge_radius() const {
		return _merge_radius;
	}


	std::vector<Candidate> candidates;
	std::vector<cv::Point3f> candidates_d;
	std::vector<cv::Point3f> candidates_pres;
	boost::mutex mutex, mutex_pres;
	std::map<int,cv::Mat > RT;
	std::vector<cv::Point3f> detections;
	std::vector<int> clients;
	std::vector<int> stored_clients;



private:
	gestoos::tracking::Point3dTracker _tracker;
	cv::Mat _detector_map;
	double _init_threshold;
	double _tracking_threshold;
	double _merge_radius;
	cv::Point2d last_show_pt;
	//Calibration options
	bool _calibrate;


	cv::Point filtered_pt;
	//	UdpSocket sock;
	UdpTransmitSocket* toRenderSocket;
	std::string ip;
	int port;

	std::vector<ShowLabel *> labels;

	void _check_redundant_candidates();
	void _check_redundant_candidates_pres();

	bool _check_synchronous_receive();
	double _processing_period_ms; //ms between two consecutive "processing frames"
	gestoos::Timestamp _last_received;

};



class ThreadedTracker
{
public:
	ThreadedTracker(){	gc=NULL; }
	~ThreadedTracker(){}

	void operator()()
	{
		int key=0;
		while(true)
		{
			if(!gc) continue;

			gc->update();

		/*	pcl::visualization::PCLVisualizer viewer("test");

			//pcl::visualization::PointCloudColorHandlerCustom<pcl::PointXYZRGB> c01_icp (cloud01_icp_ptr, 255, 255, 255);

			viewer.addPointCloud(gc->cloud, "cloud"); //red
			//  viewer.addPointCloud(cloud05, "cloud05");
			//  viewer.addPointCloud(cloud01_RT, c01_RT, "cloud01_RT");


			viewer.addCoordinateSystem(1.0, "cloud", 0);
			viewer.setBackgroundColor(0.05, 0.05, 0.05, 0); // Setting background to a dark grey
			//      viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 7, "test");
			//  viewer.setPointCloudRenderingProperties (pcl::visualization::PCL_VISUALIZER_POINT_SIZE, 7, "center05");

			if (!viewer.wasStopped())
			{ // Display the visualiser until 'q' key is pressed
			viewer.spinOnce();
			}
*/
		//	gc->send_hand_positions();
			//gc->send_presence_positions();
		}
	}

	GestureComposer * gc;

};

#endif /* SAMPLESSDK_GESTOOS_CLIENT_SRC_GESTURECOMPOSER_HPP_ */
