/*
 * calibrate_and_mix_points.cpp
 *
 *
 * Created on: Feb 15,2016
 * 		Author: Arantxa Casanova
 */


#include <iostream>
#include <sstream>
#include <time.h>
#include <stdio.h>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <imageplus/kinect_calibration/calibrate_RGB.hpp>

#include <pcl/visualization/cloud_viewer.h>
#include <pcl/io/oni_grabber.h>
#include <boost/thread/thread.hpp>
#include <pcl/common/common_headers.h>
#include <pcl/io/pcd_io.h>
#include <pcl/visualization/pcl_visualizer.h>
#include <pcl/console/parse.h>
#include <pcl/common/transforms.h>
#include <pcl/filters/statistical_outlier_removal.h>
#include <pcl/io/ply_io.h>
#include <pcl/filters/filter.h>
#include <pcl/point_types.h>
#include <imageplus/toolbox/openni.hpp>

#ifndef _CRT_SECURE_NO_WARNINGS
# define _CRT_SECURE_NO_WARNINGS
#endif

using namespace cv;
using namespace std;
using namespace imageplus;
using namespace pcl;

void get3Dpoint_from_pixels();
void test();
void intersection_pyrmidalCut_svd(pcl::PointXYZRGB &point, Mat& RT);
void intersection_pyrmidalCut_extrinsics(
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT, Mat& extr_05,
		Mat& extr_01, const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud05_cut,
		const pcl::PointCloud<pcl::PointXYZRGB>::Ptr &cloud01_RT_cut);

int main(int argc, char* argv[]) {

	const string inputSettingsFile = argc > 1 ? argv[1] : "default.xml";
	const string op = argc > 2 ? argv[2] : "default.xml";

	//obtaining parameters

	std::string in01_file, in05_file, inRGB01_file, inRGB05_file, ex01_file,
			ex05_file, manual_file, detected_fingers01, detected_fingers05,
			depth01_file, depth05_file, image01_file, image05_file,
			config_file_ext01, config_file_ext05;
	vector<Point2f> annoted_fingers01, annoted_fingers05;
	vector<std::string> img_list01, img_list05;

	FileStorage in(inputSettingsFile, FileStorage::READ);

	if (!in.isOpened())
		cout << "Could not open MAIN configuration file: \\" << endl;

	in["in05_file"] >> in05_file;
	in["in01_file"] >> in01_file;
	in["inRGB01_file"] >> inRGB01_file;
	in["inRGB05_file"] >> inRGB05_file;
	in["ex01_file"] >> ex01_file;
	in["ex05_file"] >> ex05_file;
	in["depth01_file"] >> depth01_file;
	in["depth05_file"] >> depth05_file;
	in["image01_file"] >> image01_file;
	in["image05_file"] >> image05_file;
	in["manual_file"] >> manual_file;
	in["detected_fingers01"] >> detected_fingers01;
	in["detected_fingers05"] >> detected_fingers05;
	in["config_file_ext01"] >> config_file_ext01;
	in["config_file_ext05"] >> config_file_ext05;
	in.release();

	FileStorage m(manual_file, FileStorage::READ);

	if (!m.isOpened())
		cout << "Could not open MAIN configuration file: \\" << endl;

	m["annoted_fingers01"] >> annoted_fingers01;
	m["annoted_fingers05"] >> annoted_fingers05;
	m["img_list01"] >> img_list01;
	m["img_list05"] >> img_list05;
	m.release();
	// choosing option
	if (argc == 3 && op == "calc_ext") { //calculates extrinsics

		kinect_calibration::getCalibrationParameters(config_file_ext01);

		kinect_calibration::getCalibrationParameters(config_file_ext05);

	} else if (argc == 3 && op == "ext") { //registration using extrinsics
		kinect_register_reconstruction::scene_reconstruction_ownPC(in01_file,in05_file, inRGB01_file, inRGB05_file, ex01_file, ex05_file,depth01_file, depth05_file, image01_file, image05_file);
	} else if (argc == 3 && op == "fingers_ICP") { //registration using manual detected fingers + ICP
		application::fingers_RT(in01_file, in05_file, ex05_file, inRGB01_file,inRGB05_file, img_list01, img_list05, annoted_fingers01,annoted_fingers05, depth01_file, depth05_file, image01_file,image05_file, 1);
	} else if (argc == 3 && op == "fingers_SVD") { //registration using manual detected fingers + SVD
		application::fingers_RT(in01_file, in05_file, ex05_file, inRGB01_file,inRGB05_file, img_list01, img_list05, annoted_fingers01,annoted_fingers05, depth01_file, depth05_file, image01_file,image05_file, 0);
	} else if (argc == 3 && op == "fingers_detector") { //registration using a finger detector
		application::fingers_RT_detector(in01_file, in05_file, ex05_file,inRGB01_file, inRGB05_file, detected_fingers01,detected_fingers05, depth01_file, depth05_file, image01_file,image05_file);
	} else if (argc == 3 && op == "test") { //print application output

		cout << "Visualizing tracker point cloud " << endl;
		vector < cv::Point3f > detections;
		vector<int> clients;
		FileStorage m("/imatge/acasanova/work/test_demo.xml",
				FileStorage::READ);

		if (!m.isOpened())
			cout << "Could not open file: \\" << endl;

		m["Detections"] >> detections;
		m["Clients"] >> clients;
		m.release();

		pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud01(
				new pcl::PointCloud<pcl::PointXYZRGB>);

		pcl::PointXYZRGB point_pcl;

		pcl::visualization::PCLVisualizer viewer("test");
		//	 viewer.addCoordinateSystem(1.0, "cloud", 0);
		viewer.setBackgroundColor(0.05, 0.05, 0.05, 0);
		for (unsigned int i = 0; i < detections.size(); i++) {

			point_pcl.x = detections[i].x;
			point_pcl.y = detections[i].y;
			point_pcl.z = detections[i].z;
			if (clients[i] == 1) {
				point_pcl.r = 0;
				point_pcl.g = 255;
				point_pcl.b = 0;
			} else if (clients[i] == 0) {
				point_pcl.r = 0;
				point_pcl.g = 0;
				point_pcl.b = 255;
			} else {
				point_pcl.r = 255;
				point_pcl.g = 0;
				point_pcl.b = 0;
			}
			boost::this_thread::sleep(boost::posix_time::milliseconds(100));
			cloud01->points.push_back(point_pcl);
			if (!viewer.updatePointCloud(cloud01, "cloud"))
				viewer.addPointCloud(cloud01, "cloud");
			if (!viewer.wasStopped())
				viewer.spinOnce();
		}
		cout << "Printing " << cloud01->points.size() << " points done!"
				<< endl;

		std::vector<int> indices;
		pcl::removeNaNFromPointCloud(*cloud01, *cloud01, indices);
		pcl::io::savePLYFile("/imatge/acasanova/share/APP_CLOUD.ply", *cloud01);

		while (!viewer.wasStopped()) { // Display the visualiser until 'q' key is pressed
			viewer.spinOnce();
		}

	} else {
		cout << "No option selected " << endl;
	}

	return 0;
}

